/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/* Cherokee
 *
 * Authors:
 *      Alvaro Lopez Ortega <alvaro@alobbs.com>
 *
 * Copyright (C) 2001, 2002, 2003, 2004 Alvaro Lopez Ortega
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "common.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>

#include "mmap.h"
#include "list_ext.h"


#ifndef MAP_FILE
# define MAP_OPTIONS MAP_SHARED
#else
# define MAP_OPTIONS MAP_SHARED | MAP_FILE
#endif


typedef struct {
	struct list_head base;
	
	dev_t  dev;
	ino_t  ino;
	size_t size;

	char  *mmap;

	int    use_count;
} cherokee_mmap_entry_t ;

#define MMAP_ENTRY(e) ((cherokee_mmap_entry_t *)(e))


ret_t 
cherokee_mmap_new  (cherokee_mmap_t **mmap)
{
	int i;
	CHEROKEE_NEW_STRUCT(n,mmap);

	/* Init
	 */
	for (i=0; i<MMAP_HASH_SIZE; i++) {
		INIT_LIST_HEAD(&n->hash[i]);
	}

	CHEROKEE_RWLOCK_INIT (&n->rwlock, NULL);
	
	/* Return it
	 */
	*mmap = n;
	   
	return ret_ok;
}


static cherokee_mmap_t *__mmap_global = NULL;

ret_t 
cherokee_mmap_get_default (cherokee_mmap_t **mmap)
{
	ret_t ret;

	/* Is it exits, return an existing object
	 */
	if (__mmap_global != NULL) {
		*mmap = __mmap_global;
		return ret_ok;
	}
	
	return cherokee_mmap_init (mmap);
}


ret_t cherokee_mmap_init (cherokee_mmap_t **mmap)
{
	ret_t ret;
	cherokee_mmap_t *n;

	/* Create the new object
	 */
	ret = cherokee_mmap_new (&n);
	if (ret < ret_ok) return ret;

	/* Return the object
	 */
	__mmap_global = n;
	*mmap         = n;

	return ret_ok;
}


static void
free_item (void *item)
{
	if (MMAP_ENTRY(item)->mmap != NULL) 
	{
		munmap (MMAP_ENTRY(item)->mmap, 
			MMAP_ENTRY(item)->size);

		free (item);
	}
}


ret_t 
cherokee_mmap_free (cherokee_mmap_t *cmmap)
{
	int i;

	for (i=0; i<MMAP_HASH_SIZE; i++) {
		cherokee_list_free (&cmmap->hash[i], free_item);
	}
	
	CHEROKEE_RWLOCK_DESTROY (&cmmap->rwlock);

	free (mmap);
	return ret_ok;
}


static void
maybe_clean_up_table (cherokee_mmap_t *cmmap)
{
	int i;
	list_t *a, *b;
	int                    worse_usage = INT_MAX;
	cherokee_mmap_entry_t *worse;

	if (cmmap->items < MAX_MMAP_TABLE_ITEMS) {
		return;
	}

	for (i=0; i<MMAP_HASH_SIZE; i++) {
		list_for_each_safe (a,b, &cmmap->hash[i]) {
			/* Store the worse one because we are going to
			 * free it after the loop.
			 */
			if (worse_usage > MMAP_ENTRY(a)->use_count) {
				worse       = MMAP_ENTRY(a);
				worse_usage = worse->use_count;
			}
			
			/* Set the counter to zero.
			 */
			MMAP_ENTRY(a)->use_count = 0;
		}
	}
	
	list_del ((list_t *)worse);
	free_item (worse);
}

static void
maybe_clean_up_table_AVERAGE (cherokee_mmap_t *cmmap)
{
	list_t *a, *b;
	unsigned long i, average = 0;
	cherokee_mmap_entry_t *entry;
	
	if (cmmap->items < MAX_MMAP_TABLE_ITEMS) {
		return;
	}

	/* Calculate the average
	 */
	for (i=0; i<MMAP_HASH_SIZE; i++) {
		list_for_each (a, &cmmap->hash[i]) {
			average += MMAP_ENTRY(a)->use_count;
		}
	}
	
	average /= i;

	/* Delete olders
	 */
	for (i=0; i<MMAP_HASH_SIZE; i++) {
		list_for_each_safe (a,b, &cmmap->hash[i]) {
			entry = MMAP_ENTRY(a);
			
			if (entry->use_count < average) {
				list_del (a);
				free_item(entry);
			} else {
				entry->use_count = 0;
			}
		}
	}
}


static cherokee_mmap_entry_t *
find_hash (cherokee_mmap_t *cmmap, struct stat *s)
{
	list_t *i;

	list_for_each (i, &cmmap->hash[MMAP_HASH_FUNC(s)]) {

		if ((s->st_dev  == MMAP_ENTRY(i)->dev) &&
		    (s->st_ino  == MMAP_ENTRY(i)->ino) &&
		    (s->st_size == MMAP_ENTRY(i)->size))
		{
			MMAP_ENTRY(i)->use_count++;
			return MMAP_ENTRY(i);
		}
	}

	return NULL;
}

static cherokee_mmap_entry_t *
find (cherokee_mmap_t *cmmap, int fd, struct stat *s) 
{
	cherokee_mmap_entry_t *entry;
	void *area;

	CHEROKEE_RWLOCK_READER (&cmmap->rwlock);  /* 1.- Adquire */
	entry = find_hash (cmmap, s);
	CHEROKEE_RWLOCK_UNLOCK (&cmmap->rwlock);  /* 2.- Release */

	if (entry != NULL) {
		return entry;
	}

	area = mmap (NULL,           /* void   *start  */
		     s->st_size,     /* size_t  length */
		     PROT_READ,      /* int     prot   */
		     MAP_OPTIONS,    /* int     flag   */
		     fd,             /* int     fd     */
		     0);             /* off_t   offset */

	if (area == MAP_FAILED) {
		return NULL;
	}

	{
		cherokee_mmap_entry_t *n;
		n = (cherokee_mmap_entry_t *) malloc (sizeof(cherokee_mmap_entry_t));
		if (n == NULL) return NULL;

		INIT_LIST_HEAD((list_t*)n);

		n->dev       = s->st_dev;
		n->ino       = s->st_ino;
		n->size      = s->st_size;
		n->mmap      = area;
		n->use_count = 0;
		
		CHEROKEE_RWLOCK_WRITER (&cmmap->rwlock);    /* 1.- Adquire */

		maybe_clean_up_table (cmmap);
		list_add ((list_t *)n, &cmmap->hash[MMAP_HASH_FUNC(s)]);

		CHEROKEE_RWLOCK_UNLOCK (&cmmap->rwlock);    /* 2.- Release */

		return n;
	}

	return NULL;
}


ret_t
cherokee_mmap_get (cherokee_mmap_t *cmmap, const char *filename, void **ret, size_t *ret_len)
{
	int fd;
	struct stat s;
	cherokee_mmap_entry_t *entry;

	fd = open (filename, O_RDONLY);
	if (fd == -1) {
		return ret_error;
	}

	fstat(fd, &s);
	if (S_ISDIR(s.st_mode)) {
		return ret_error;
	}

	entry = find (cmmap, fd, &s);
	close(fd);

	if (entry == NULL) {
		return ret_error;
	}

	*ret     = entry->mmap;
	*ret_len = entry->size;
	return ret_ok;
}


ret_t
cherokee_mmap_get_fast (cherokee_mmap_t *cmmap, int fd, struct stat *s, void **ret, size_t *ret_len)
{
	cherokee_mmap_entry_t *entry;

	entry = find (cmmap, fd, s);
	if (entry == NULL) {
		return ret_error;
	}

	*ret     = entry->mmap;
	*ret_len = entry->size;

	return ret_ok;
}
